#include <math.h>
#include <stdio.h>
#include "individl.h"

#ifdef D1


//============     ********    D1    ********     ===========================


void genotype::rnd_init() {
  gen=RaNdOm(16384);
  gen<<=14;gen+=RaNdOm(16384);gen<<=4;
  gen+=RaNdOm(16);
  if(bitlength!=32) gen&=(1UL<<bitlength)-1;
};

void genotype::operator*(genotype &e) {
  signed char k;
  unsigned long e1,e2,maszk;
  k=RaNdOm(bitlength);
  maszk=(1UL<<k)-1;
  e1 = maszk & gen; // e1-ben gen jobboldala
  gen &= ~maszk;    // gen-ben marad a baloldal
  e2 = maszk & e.gen;
  e.gen &= ~maszk;
  gen |= e2;
  e.gen |= e1;
};

signed char genotype::operator*(unsigned int P_m) {
// P_m is for 1 bit but we cheat a little...
  signed char voltmut=0;
  unsigned int  k=RaNdOm( (P_m+(bitlength/2))/bitlength );
  unsigned long j;
  if(k==0) {
    voltmut++;
    k=RaNdOm(bitlength);
    j=1<<k;
    if(gen == (gen | j)) gen &= ~j;
    else gen |= j; };
  return voltmut;
};

genotype::genotype(GENOTYPE g) { gen=g; };

genotype::genotype(const genotype& g) { gen=g.gen; };

genotype::~genotype() {;};

genotype& genotype::operator=(const genotype& g) {
  gen=g.gen;
  return *this;
};

signed char genotype::operator==(genotype& g) { return gen==g.gen; };

//===========================================================================

fenotype::fenotype(fenotype& o,double r) {
  double d=((double)rand())/RAND_MAX*r;
  fen=o.fen+(RaNdOm(2)==0 ? -d : d);
};

fenotype::fenotype(fenotype& f1,fenotype& f2) {
  fen=f1.fen+(double(rand())/RAND_MAX)*(f2.fen-f1.fen);
};

fenotype::~fenotype() {;};

signed char fenotype::between(fenotype& f1,fenotype& f2) {
  if(f1.fen<=f2.fen) return f1.fen<=fen && fen<=f2.fen;
  return f2.fen<=fen && fen<=f1.fen;
};

double fenotype::operator-(fenotype& f) {
  return fen>f.fen ? fen-f.fen : f.fen-fen;
};

fenotype::fenotype(FENOTYPE f) { fen=f; };

fenotype::fenotype(const fenotype& f) { fen=f.fen; };

fenotype& fenotype::operator=(const fenotype& f) {
  fen=f.fen;
  return *this;
};

//===========================================================================

ostream& operator<<(ostream& os,genotype& g) {
  signed char s[10];
  sprintf(s,"%8lX",g.gen);
  os << s;
  return os;
};

ostream& operator<<(ostream& os,fenotype& f) {
  os << f.fen;
  return os;
};

#else


//============     ********    D2+   ********     ===========================


void genotype::rnd_init() {
  for(signed char i=0;i<DIM;++i) {
    gen[i]=RaNdOm(16384);
    gen[i]<<=14;gen[i]+=RaNdOm(16384);gen[i]<<=4;
    gen[i]+=RaNdOm(16);
    if(bitlength!=32) gen[i]&=(1UL<<bitlength)-1;
  };
};

void genotype::operator*(genotype &e) {
  signed char   k;
  unsigned long e1,e2,maszk;
  for(signed char i=0;i<DIM;++i) {
    k=RaNdOm(bitlength);
    maszk=(1UL<<k)-1;
    e1 = maszk & gen[i]; // e1-ben gen jobboldala
    gen[i] &= ~maszk;    // gen-ben marad a baloldal
    e2 = maszk & e.gen[i];
    e.gen[i] &= ~maszk;
    gen[i] |= e2;
    e.gen[i] |= e1; };
};

signed char genotype::operator*(unsigned int P_m) {
// P_m is for 1 bit but we cheat a little...
  signed char   voltmut=0;
  unsigned int  p_m=(P_m+(bitlength/2))/bitlength,
		k;
  unsigned long j;
  for(int i=0;i<DIM;++i) {
    k=RaNdOm(p_m);
    if(k==0) {
      voltmut++;
      k=RaNdOm(bitlength);
      j=1<<k;
      if(gen[i] == (gen[i] | j)) gen[i] &= ~j;
      else gen[i] |= j; }; };
  return voltmut;
};

genotype::genotype(GENOTYPE g) {
  for(signed char i=0;i<DIM;++i) gen[i]=g[i];
};

genotype::genotype(const genotype& g) {
  for(signed char i=0;i<DIM;++i) gen[i]=g.gen[i];
};

genotype::~genotype() {;};

genotype& genotype::operator=(const genotype& g) {
  for(signed char i=0;i<DIM;++i) gen[i]=g.gen[i];
  return *this;
};

signed char genotype::operator==(genotype& g) {
  signed char i;
  for(i=0;i<DIM && gen[i]==g.gen[i];++i);
  return i==DIM;
};

//===========================================================================

fenotype::fenotype(fenotype& o,double r) {
  double d;
  r/=sqrtDIM;
  for(signed char i=0;i<DIM;++i) {
    d=((double)rand())/RAND_MAX*r;
    fen[i]=o.fen[i]+(RaNdOm(2)==0 ? -d : d);
  };
};

fenotype::fenotype(fenotype& f1,fenotype& f2) {
  double p=double(rand())/RAND_MAX;
  for(signed char i=0;i<DIM;++i)
    fen[i]=f1.fen[i]+p*(f2.fen[i]-f1.fen[i]);
};

fenotype::~fenotype() {;};

double fenotype::operator-(fenotype& f) {
  double d=0.0;
  for(signed char i=0;i<DIM;++i) d+=(fen[i]-f.fen[i])*(fen[i]-f.fen[i]);
  return sqrt(d);
};

signed char fenotype::between(fenotype& f1,fenotype& f2) {
  fenotype o;
  for(signed char i=0;i<DIM;++i) o.fen[i]=(f1.fen[i]+f2.fen[i])/2;
  return o-*this<=o-f1;
};

fenotype::fenotype(FENOTYPE f) {
  for(signed char i=0;i<DIM;++i) fen[i]=f[i];
};

fenotype::fenotype(const fenotype& f) {
  for(signed char i=0;i<DIM;++i) fen[i]=f.fen[i];
};

fenotype& fenotype::operator=(const fenotype& f) {
  for(signed char i=0;i<DIM;++i) fen[i]=f.fen[i];
  return *this;
};

//===========================================================================

ostream& operator<<(ostream& os,genotype& g) {
  signed char s[10],i=0,j=0;
  os << "(";
  while(8*j+i<DIM) {
    sprintf(s,"%8lX",g.gen[8*j+(i=0)]);
    os << s;
    for(signed char i=1;i<8 && 8*j+i<DIM;++i) {
      sprintf(s,"%8lX",g.gen[i]);
      os << "," << s; };
    if(8*++j<DIM && i==8) os << "\n ";
  };
  os << ")";
  return os;
};

ostream& operator<<(ostream& os,fenotype& f) {
  for(int i=0; i<DIM; ) {
    os << f.fen[i];
    if(++i != DIM) os << " ";
  };
  return os;
};

#endif

