#include <math.h>
#include "individl.h"

task::task(double* param) : parameter_(param) {; };

void task::new_parameter(double* p) { parameter_=p; };

//---------------------------------------------------------------------------

#ifdef D1


//============     ********    D1    ********     ===========================


fitnesst task::fit(fenotype& fen) {
// parameter_[2] a fv sorszama
  double x=fen.fen,
  	 a[10]={ 3.04,1.098,0.674,3.537,6.173,8.679,4.503,3.328,6.937,0.7 },
	 k[10]={ 2.983,2.378,2.439,1.168,2.406,1.236,2.868,1.378,2.348,2.268 },
	 c[10]={ .192,.14,.127,.132,.125,.189,.187,.171,.188,.176 },
	 y=0.0;

  switch((int)parameter_[2]) {

    case 0 :
      return (int((x-M_PI_2)/M_PI)%2 == 0 ? sin(x)*x*0.1 : cos(x+M_PI_2)*x*0.1) +
      double(RaNdOm(10000))/10000;

    case 1 : return sin(x);

    case 2 : return fabs(x-int(x)-.5)*3+sin(x)*x;

    case 3 :
      // Han (Hansen) fv.  Ran : [0,3]
      return (((-93*x+276)*x-303)*x+142)*x-24;

    case 4 :
      return -(sin(x)+sin(10*x/3)+log(x)-0.84*x); // [2.7,7.5]

    case 5 :  return -(x+sin(x))*exp(-x*x); // [-10,10]

    case 6 :
      for(signed char j=0;j<10;++j) y+=1/(k[j]*(x-a[j])*k[j]*(x-a[j])+c[j]);
      return y; // [0,10]

    case 7 :
      // Shubert fv.  Ran : [-10,10]
      y=0.0;
      for(j=1;j<6;++j) y+=j*sin((j+1)*x+j);
      return y;

    case 8 :
      y=parameter_[3]; // delta
      #define LOG2(z)  log(z)/log(2)
      #define H2(z)  (-(z)*LOG2(z)-(1-(z))*LOG2(1-(z)))
      #define g(z)  H2( (1-sqrt( 1-(z)>=0 ? 1-(z) : 0 ))/2 )
      return -(1+g(x*x)-g(x*x+2*y*x+2*y));

    case 9 : return sin(x)*x;

    case 10 :
      // Levy 1 ; [-4,4]
      x*=x;
      return ((-250*x-27)*x+15)*x-1;

    default : return 1;
  };
};

//===========================================================================

fenotype task::fen(genotype& gen) {
// parameter_[0] : fenotypeus also hatar
// parameter_[1] : fenotypeus felso hatar

  // Gray-dekodolas

  fenotype       eredmeny;
  unsigned long bitj,g=0UL,maszk=1UL<<bitlength-1;
  g=maszk & gen.gen;                          // felso bit valtozatlan
  for(char j=bitlength-2;j>=0;--j,maszk>>=1) {
    bitj=(maszk>>1) & gen.gen;
    if((maszk & g) == 0) g|=bitj;             // valtozatlan
    else if(bitj == 0) g|=(maszk>>1); };      // atfordul


  // fenotype szamolas

  unsigned long k=(bitlength != 32 ? (1UL<<bitlength)-1 : ~0UL);
  eredmeny.fen=(parameter_[1]-parameter_[0])*g/k+parameter_[0];
  return eredmeny;
};

//===========================================================================

genotype task::invfen(fenotype& fen) {
// parameter_[0] : fenotypeus also hatar
// parameter_[1] : fenotypeus felso hatar

  // egy termeszetellenes konverzio a modellezes oltaran

  genotype eredmeny;
  double fen_=fen.fen;
  if(fen_ < parameter_[0]) fen_=parameter_[0];
  else if(parameter_[1] < fen_) fen_=parameter_[1];
  unsigned long
    k=(bitlength != 32 ? (1UL<<bitlength)-1 : ~0UL),
    g=(unsigned long)(k*(fen_-parameter_[0])/(parameter_[1]-parameter_[0])),
    gen=0UL;

  // Gray-kodolas

  unsigned long maszk,bitj;
  maszk=1UL<<bitlength-1;
  gen=maszk & g;                              // felso bit valtozatlan
  for(char j=bitlength-2;j>=0;--j,maszk>>=1) {
    bitj=(maszk>>1) & g;
    if((maszk & g) == 0) gen|=bitj;           // valtozatlan
    else if(bitj == 0) gen|=(maszk>>1); };    // atfordul
  eredmeny.gen=gen;
  return eredmeny;
};

#endif

#ifdef D2


//============     ********    D2    ********     ===========================


fitnesst task::fit(fenotype& fen) {
//parameter_[4] : fv sorszama
  double      x[2]={ fen.fen[0], fen.fen[1] };
  double      s,y;
  signed char j;

  switch((int)parameter_[4]) {

  case 0:
    if(x[0]<=10) s=-(x[0]-5)*(x[0]-5)-(x[1]-10)*(x[1]-10);
    else s=-(x[0]-15)*(x[0]-15)-(x[1]-10)*(x[1]-10);
    return s;

  case 1:
    // Branin fv. Ran : [ (-5,10) , (0,15) ]
    s=M_1_PI*5*x[0]-1.275*M_1_PI*M_1_PI*x[0]*x[0]+x[1]-6;
    s=s*s+(10-1.25*M_1_PI)*cos(x[0]);
    return -s;

  case 2:
    // Six-hump-cammel-back fv. Ran : [ (-2.5,2.5) , (-2.5,2.5) ]
    double p[3]={ x[0]*x[0],x[1]*x[1],0 };
    p[2]=p[0]*(4 - 2.1*p[0] + 0.333333333333333333333*p[0]*p[0]) +
	 x[0]*x[1] - 4*p[1]*(1 - p[1]);
    return -p[2];

  case 3:
    // Three hump camel back [?:?] opt :[0,0]
    s=x[1]*x[1];
    return -(12*s-6.3*s*s+s*s*s+6*x[2]*(x[2]+x[1]))/6;

  case 4:
    // Levy 3 [-10,10]^2
    y=0.0;
    for(j=1;j<6;++j) y+=j*cos((j-1)*x[0]+j);
    s=0.0;
    for(j=1;j<6;++j) s+=j*cos((j+1)*x[1]+j);
    return -y*s;

  case 5:
    // Levy 13 [-10,10]^2
    y=sin(3*M_PI*x[0]);
    y*=y;
    s=y+(x[0]-1)*(x[0]-1)*(1+sin(3*M_PI*x[1])*sin(3*M_PI*x[1]));
    s+=(x[1]-1)*(1+sin(2*M_PI*x[1])*sin(2*M_PI*x[1]));
    return -s;

  case 6:
    // Beale func.   opt [3, .5] 0
    return -( (1.5 -x[0]+x[0]*x[1])*(1.5 -x[0]+x[0]*x[1])+
	      (2.25-x[0]+x[0]*x[1]*x[1])*(2.25-x[0]+x[0]*x[1]*x[1]) +
	      (2.625-x[0]+x[0]*x[1]*x[1]*x[1])*
	      (2.625-x[0]+x[0]*x[1]*x[1]*x[1]));

  case 7:
    // Booth problem opt [1,3] 0
    return -( (x[0]+2*x[1]-7)*(x[0]+2*x[1]-7)+
	      (2*x[0]+x[1]-5)*(2*x[0]+x[1]-5) );

  default:
    return -(x[0]-10)*(x[0]-10)-(x[1]-10)*(x[1]-10);
  };
};

#endif

#ifdef D3


//============     ********    D3    ********     ===========================


fitnesst task::fit(fenotype& fen) {
// parameter_[6] fv sorszama
  double x[3]={ fen.fen[0],fen.fen[1],fen.fen[2] };
  double d,e;

  switch((int)parameter_[6]) {

  default :
    d=(1-x[0]);d*=d*(x[0]*x[0]+1);
    e=x[0]-x[1]*x[1];e*=e;e+=(x[1]-1)*(x[1]-1);d+=e;
    e=x[0]-x[2]*x[2];e*=e;e+=(x[2]-1)*(x[2]-1);
    return -(d+e); // -10 < x[i] < 10
  };
};

#endif

#ifdef D10


//============     ********    D10   ********     ===========================


fitnesst task::fit(fenotype& fen) {
// parameter_[20] fv sorszama
  double e,d=0.0;

  switch((int)parameter_[20]) {

  default :
    for(char i=0;i<10;++i) {
      e=fen.fen[i]*fen.fen[i];e*=e*fen.fen[i];e*=e;d+=e; };
    return -d; // -1 < x[i] < 2
  };
};

#endif

#ifndef D1


//============     ********    D2+   ********     ===========================


fenotype task::fen(genotype& gen) {
// parameter_[0]...[DIM-1] : alsoh
// parameter_[DIM]...[2*DIM-1] : felsoh

  fenotype eredmeny;

  // Gray-dekodolas

  unsigned long g[DIM],maszk,bitj;
  for(char c=0;c<DIM;++c) {
    g[c]=0;
    maszk=1UL<<bitlength-1;
    g[c]=maszk & gen.gen[c]; // felso bit valtozatlan
    for(char j=bitlength-2;j>=0;--j,maszk>>=1) {
      bitj=(maszk>>1) & gen.gen[c];
      if((maszk & g[c]) == 0) g[c]|=bitj;     // valtozatlan
      else if(bitj == 0) g[c]|=(maszk>>1);    // atfordul
    }; };

  // fenotype szamolas

  unsigned long k=(bitlength != 32 ? (1UL<<bitlength)-1 : ~0UL);
  for(c=0;c<DIM;++c)
    eredmeny.fen[c]=(parameter_[DIM+c]-parameter_[c])*g[c]/k+parameter_[c];
  return eredmeny;
};

//===========================================================================

genotype task::invfen(fenotype& fen_) {
// egy termeszetellenes konverzio a modellezes oltaran
// parameter_[0]...[DIM-1] : alsoh
// parameter_[DIM]...[2*DIM-1] : felsoh

  double        h[DIM],fen[DIM];
  for(char c=0;c<DIM;++c) {
    h[c]=parameter_[DIM+c]-parameter_[c];
    if(fen_.fen[c] < parameter_[c])
      fen[c]=parameter_[c];
    else
      if(parameter_[DIM+c] < fen_.fen[c])
	fen[c]=parameter_[DIM+c];
      else
	fen[c]=fen_.fen[c]; };
  unsigned long g[DIM],k=(bitlength != 32 ? (1UL<<bitlength)-1 : ~0UL);
  genotype      gen;
  for(c=0;c<DIM;++c) g[c]=(unsigned long)(k*(fen[c]-parameter_[c])/h[c]);

  // Gray-kodolas

  unsigned long maszk,bitj;
  for(c=0;c<DIM;++c) {
    gen.gen[c]=0;
    maszk=1UL<<bitlength-1;
    gen.gen[c]=maszk & g[c]; // felso bit valtozatlan
    for(char j=bitlength-2;j>=0;--j,maszk>>=1) {
      bitj=(maszk>>1) & g[c];
      if((maszk & g[c]) == 0) gen.gen[c]|=bitj;     // valtozatlan
      else if(bitj == 0) gen.gen[c]|=(maszk>>1);    // atfordul
    }; };
  return gen;
};

#endif
