#include "individl.h"

void genotype::rnd_init() {
  for(int i=0;i<GT_LENGTH;++i) {
    gen[i]=RaNdOm((int)UCHAR_MAX+1);
  };
};

void genotype::operator*(genotype &e) {
  int           kk=RaNdOm(BITLENGTH);
  unsigned char e1,e2,maszk;
  int           i=kk/CHAR_BIT,k=kk%CHAR_BIT;
  maszk=(1<<k)-1;
  e1 = maszk & gen[i];   // e1-ben gen jobboldala
  gen[i] &= ~maszk;      // gen-ben marad a baloldal
  e2 = maszk & e.gen[i];
  e.gen[i] &= ~maszk;
  gen[i] |= e2;
  e.gen[i] |= e1;
  for(--i; i>=0; --i) {
    e1=gen[i]; gen[i]=e.gen[i]; e.gen[i]=e1;
  };
};

signed char genotype::operator*(unsigned int P_m) {
// P_m is for 1 bit but we cheat a little...
  signed char   voltmut=0;
  unsigned int  p_m=(P_m+(CHAR_BIT/2))/CHAR_BIT,
		k;
  unsigned char j;
  for(int i=0;i<GT_LENGTH;++i) {
    k=RaNdOm(p_m);
    if(k==0) {
      voltmut++;
      k=RaNdOm(CHAR_BIT);
      j=1<<k;
      if(gen[i] == (gen[i] | j)) gen[i] &= ~j;
      else gen[i] |= j; }; };
  return voltmut;
};

genotype::genotype(GENOTYPE g) {
  for(int i=0;i<GT_LENGTH;++i) gen[i]=g[i];
};

genotype::genotype(const genotype& g) {
  for(int i=0;i<GT_LENGTH;++i) gen[i]=g.gen[i];
};

genotype::~genotype() {;};

genotype& genotype::operator=(const genotype& g) {
  for(int i=0;i<GT_LENGTH;++i) gen[i]=g.gen[i];
  return *this;
};

signed char genotype::operator==(genotype& g) {
  for(int i=0;i<GT_LENGTH && gen[i]==g.gen[i];++i);
  return i==GT_LENGTH;
};

//===========================================================================

fenotype::fenotype(fenotype& o,double r) {
// we want o-*this<=r here
  int           rr=RaNdOm((int)r)+1,
		kk,k,i;
  unsigned char m;
  *this=o;
  for(int j=0; j<rr; ++j) {
    kk=RaNdOm(BITLENGTH);
    i=kk/CHAR_BIT;
    k=kk%CHAR_BIT;
    m=1<<k;
    if(fen[i] == (fen[i] | m)) fen[i] &= ~m;
    else fen[i] |= m;
  };
};

fenotype::fenotype(fenotype& f1,fenotype& f2) {
// we want f1-f2=(*this-f1)+(*this-f2) here
  unsigned char xor,j;
  for(int i=0; i<GT_LENGTH; ++i) {
    xor = f1.fen[i] ^ f2.fen[i];
    j=RaNdOm(CHAR_BIT);
    fen[i] = (j & xor) | (~xor & f1.fen[i]);
  };
};

fenotype::~fenotype() {;};

double fenotype::operator-(fenotype& f) {
  unsigned char c;
  int           d=0,j,i=0;
  for(; i<GT_LENGTH-1; ++i) {
    c = fen[i] ^ f.fen[i];
    for(j=0; j<CHAR_BIT; c>>=1,++j)
      if(c & 1) ++d;
  };
  c = fen[i] ^ f.fen[i];
  for(j=0; j<LAST_LENGTH; c>>=1,++j)
    if(c & 1) ++d;
  return (double)d;
};

signed char fenotype::between(fenotype& f1,fenotype& f2) {
  double d1=f1-f2,
	 d2=*this-f1,
	 d3=*this-f2;
  return (d2+d3) < 1.4*d1;
};

fenotype::fenotype(FENOTYPE f) {
  for(int i=0;i<GT_LENGTH;++i) fen[i]=f[i];
};

fenotype::fenotype(const fenotype& f) {
  for(int i=0;i<GT_LENGTH;++i) fen[i]=f.fen[i];
};

fenotype& fenotype::operator=(const fenotype& f) {
  for(int i=0;i<GT_LENGTH;++i) fen[i]=f.fen[i];
  return *this;
};

//===========================================================================

ostream& operator<<(ostream& os,genotype& g) {
  os << hex;
  int i=GT_LENGTH-1;
  if(LAST_LENGTH!=CHAR_BIT) {
    os << (int)(g.gen[i] & (((unsigned char)1 << LAST_LENGTH)-1)) ;
    if(--i != -1) os << " ";
  };
  for(; i>=0; ) {
    os << (int)g.gen[i];
    if(--i != -1) os << " ";
  };
  return os;
};

ostream& operator<<(ostream& os,fenotype& f) {
  os << hex;
  int i=GT_LENGTH-1;
  if(LAST_LENGTH!=CHAR_BIT) {
    os << (int)(f.fen[i] & (((unsigned char)1 << LAST_LENGTH)-1)) ;
    if(--i != -1) os << " ";
  };
  for(; i>=0; ) {
    os << (int)f.fen[i];
    if(--i != -1) os << " ";
  };
  return os;
};
