#include "populati.h"

void fewmem() {
  cerr << "Sorry, we have run out of memory !\n";
  exit(1);
};

// -----------------------------------------------------------------------

population::population(task& f,int esz,signed char rnd, double maxsp,
                       const individual* ia) {

  // starts with esz random individuals or fills the population with ia
  // the ia array must contain esz individuals if ia!=NULL
  // [maxsp*esz] is the maximal allowed number of species

  if(esz<2) return;    // has no meaning
  if(rnd) RaNdOmIzE;
  individual*   e=(ia==NULL ? (new individual(f)) : (new individual(ia[0])));
  int           i;
  if(e==NULL) fewmem();
  freeze=0;
  task_=&f;
  iterator=NULL;
  min_rad_index=1;
  if(maxsp<=0.0 || maxsp>=1.0) maxsp=1.0/4.0;
  max_species_number=(int)(maxsp*esz+.5);
  species_number=1;
  if((taxon_chart.lastind=
      taxon_chart.indvdls=
      taxon_chart.maxe=new ind_list_t)==NULL) fewmem();
  taxon_chart.indvdls->next=NULL;
  taxon_chart.indvdls->spec=&taxon_chart;
  Gfitlowb=taxon_chart.fitlowb=taxon_chart.fitsum=e->fit();
  taxon_chart.indvdls->e=*e;
  taxon_chart.count=taxon_chart.countall=1;
  taxon_chart.rad_index=0;
  taxon_chart.firstson=taxon_chart.brother=taxon_chart.dad=NULL;
  virt_tree_pntr=&taxon_chart;
  for(i=1;i<esz;++i) {
    delete e;
    if((e=(ia==NULL ? (new individual(f)) : (new individual(ia[i])) ))==NULL)
      fewmem();
    insert(e);
  };
};

// -----------------------------------------------------------------------

population::~population() {
  ind_list_t *eseg,*e=taxon_chart.indvdls;
  taxon_chart_t   *tseg,*t=taxon_chart.firstson;
  while(e != NULL) { eseg=e->next;delete e;e=eseg; };
  while(t != NULL) { tseg=t->brother;del_tree(t);t=tseg; };
};

// -----------------------------------------------------------------------

void population::del_tree(taxon_chart_t* t) {
  taxon_chart_t* tseg=t->firstson;
  delete t;t=tseg;
  while(t != NULL) { tseg=t->brother;del_tree(t);t=tseg; };
};

// -----------------------------------------------------------------------

void population::insert(const individual *e) {
//  inserts to virt_tree_pntr
  ind_list_t *hova=virt_tree_pntr->lastind,
             *mit=new ind_list_t,
             *seged;
  if(mit==NULL) fewmem();
  mit->e=*e;
  mit->spec=virt_tree_pntr;

  // insert

  seged=hova->next;
  hova->next=mit;
  mit->next=seged;

  // correction of fields

  taxon_chart_t *f=virt_tree_pntr;
  f->lastind=mit;
  f->fitsum+=mit->e.fit();
  if(f->maxe->e.fit() < mit->e.fit()) f->maxe=mit;
  else {
    if(mit->e.fit() < f->fitlowb) f->fitlowb=mit->e.fit();
    if(mit->e.fit() < Gfitlowb) Gfitlowb=mit->e.fit(); };
  f->count++;
  for(;f != NULL;f=f->dad) f->countall++;
};

// -----------------------------------------------------------------------

taxon_chart_t* population::insert_to_new(const individual *e) {
// virt_tree_pntr as dad gets new child
// if it wasn't possible we insert into virt_tree_pntr
// return: created species, NULL anyway
  if(freeze || species_number>=max_species_number) { // mustn't create sp.
    insert(e); return NULL; };
  ind_list_t *hova,
             *mit=new ind_list_t,
             *seged;
  if(mit==NULL) fewmem();
  taxon_chart_t    *f=new taxon_chart_t,
                   *seg=virt_tree_pntr;
  if(f==NULL) fewmem();
  mit->e=*e;
  mit->spec=f;
  f->firstson=f->brother=NULL;
  f->dad=virt_tree_pntr;
  f->lastind=f->indvdls=mit;
  f->countall=f->count=1;
  f->maxe=mit;
  f->rad_index=(f->dad->rad_index<min_rad_index ?
		min_rad_index : f->dad->rad_index+1);
  f->fitlowb=f->fitsum=e->fit();

  if(seg->firstson==NULL) { seg->firstson=f;hova=seg->lastind; }
  else {
    seg=seg->firstson;
    int i=1;
    while(seg->brother!=NULL) { seg=seg->brother,++i; };
    if(i==255) {  // too many children
      delete mit;delete f;insert(e);return NULL; };
    seg->brother=f;
    for(hova=seg->lastind;(seg=seg->firstson)!=NULL;hova=seg->lastind)
      for(;seg->brother!=NULL;seg=seg->brother);
  };

  ++species_number;

  // insert

  seged=hova->next;
  hova->next=mit;
  mit->next=seged;

  // correction of fields

  if(mit->e.fit() < Gfitlowb) Gfitlowb=mit->e.fit();
  seg=f;
  for(f=f->dad;f != NULL;f=f->dad) f->countall++;
  return seg;
};

// -----------------------------------------------------------------------

signed char population::kill(ind_list_t* t) {
  taxon_chart_t* f=t->spec;
  ind_list_t*    m=NULL;
  if(f->count == 1) return 0; // there must be min. one ind.
  if(f->maxe == t)  return 0; // best ind.
  if(f->indvdls==t && f->dad!=NULL) { f=f->dad; };
  if(f->indvdls!=t) {
    m=f->indvdls;
    while(m->next != t) { m=m->next; };
    m->next=t->next; }
  else taxon_chart.indvdls=t->next;
  f=t->spec;f->count--;
  if(f->indvdls==t) f->indvdls=m->next;
  if(f->lastind==t) f->lastind=m;
  f->fitsum-=t->e.fit();
  delete t;
  for(;f != NULL;f=f->dad) f->countall--;
  return 1;
};

// -----------------------------------------------------------------------

void population::species(taxon_chart_t* f,unsigned char* spe) {
  // 8 bytes of spe must be available
  taxon_chart_t *dad,*seged;
  int szint,i;
  szint=7;
  while((dad=f->dad) != NULL) {
    for(i=1,seged=dad->firstson;seged != f;seged=seged->brother,++i);
    spe[szint--]=i;f=dad; };
  if(szint>-1) {
    for(i=szint+1;i<8;++i) spe[i-szint-1]=spe[i];
    for(i=7-szint;i<8;++i) spe[i]=0; };
};

// -----------------------------------------------------------------------

int population::parent_level() {
  taxon_chart_t* seged=parents[0]->spec;
  int            i=0;
  while(seged->dad!=NULL) { ++i;seged=seged->dad; };
  return i;
};

// -----------------------------------------------------------------------

taxon_chart_t* population::pointer(unsigned char* spe) {
  taxon_chart_t* f=&taxon_chart;
  int       j,i=0;
  while(f!=NULL && spe[i]!=0 && i<8) {
    if((f=f->firstson) == NULL) break;
    for(j=1;j<spe[i];++j)
      if((f=f->brother) == NULL) break;
    i++; };
  return f;
};

// -----------------------------------------------------------------------

void population::iterstart(unsigned char* Faj) {
  taxon_chart_t* f=pointer(Faj);
  if((iter_tree=f) == NULL) iterator=NULL;
  else iterator=iter_tree->indvdls;
};

// -----------------------------------------------------------------------

individual* population::iteradd() {
  individual* e=&(iterator->e);
  iterator=iterator->next;
  return e;
};

// -----------------------------------------------------------------------

signed char population::iter_all_end(ind_list_t* elmut) {
  if(elmut == NULL) return 1;
  for(taxon_chart_t* f=elmut->spec;f != NULL;f=f->dad)
    if(iter_tree == f) return 0;
  return 1;
};

// -----------------------------------------------------------------------

signed char population::iterend() {
  if(iterator == NULL) return 1;
  if(iterator->spec != iter_tree) return 1;
  return 0;
};

// -----------------------------------------------------------------------

void population::put_in(const char *c) {
  // c="si" : i. parent , c="ui" : i. offspring
  // c="si+" : i. parent to new species, c="ui+" : i. offspring to new spec.
  // c="?i+?j" to new species two together
  // we loose parents but not virt_tree_pntr after puting in
  individual     &e=max_ind; // we're saving memory
  signed char    sorszam=c[1]=='1' ? 0 : 1;
  taxon_chart_t* seged=virt_tree_pntr;
  if(c[0]=='s') {
    e=parents[sorszam]->e;
    kill(parents[sorszam]); }
  else e=offspring[sorszam];
  if(c[2]!='+') insert(&e);
  else if((virt_tree_pntr=insert_to_new(&e))==NULL) virt_tree_pntr=seged;
  if(c[2]=='+' && c[3]!=0) {
    sorszam=c[4]=='1' ? 0 : 1;
    if(c[3]=='s') {
      e=parents[sorszam]->e;
      kill(parents[sorszam]); }
    else e=offspring[sorszam];
    insert(&e);
  };
  virt_tree_pntr=seged;
};

// -----------------------------------------------------------------------

void population::parent_selection() {
// relative probablity to fitlowb-s

  // species selection

  taxon_chart_t*    f;
  ind_list_t* el=taxon_chart.indvdls;
  int          i,vtlen=RaNdOm(taxon_chart.countall);
  for(i=0;i<vtlen;++i) el=el->next;
  f=el->spec;

  // first parent

  fitnesst osszfit=f->fitsum-f->count*f->fitlowb; //relative sum to fitlowb-s
  fitnesst seged=(el=f->indvdls)->e.fit()-f->fitlowb; // first individual
  vtlen=RaNdOm(RAND_MAX-1)+1;
  fitnesst fvtlen=(osszfit/RAND_MAX)*vtlen;
  for(;seged<fvtlen && el->next->spec==f;seged+=el->e.fit()-f->fitlowb)
    el=el->next;
  parents[0]=el;

  // second parent

  i=0;
  do {
    seged=(el=f->indvdls)->e.fit()-f->fitlowb; // rel. fitness of 1. ind.
    vtlen=RaNdOm(RAND_MAX-1)+1;
    fvtlen=(osszfit/RAND_MAX)*vtlen;
    for(;seged<fvtlen && el->next->spec==f;seged+=el->e.fit()-f->fitlowb)
      el=el->next;
    i++; }
  while (i<3 && el->e==parents[0]->e); // if the same try to choose another
  parents[1]=el;
};

// -----------------------------------------------------------------------

void population::dying_off(unsigned int ptovabb,int db) {
// examining about every ptovabb-th ind. kills one ind. db times.
// fitness is 'seen' as (fitness-Gfitlowb)/[ind_list_t].spec->count
  if(db<=0) return;
  if(ptovabb>taxon_chart.countall) ptovabb=taxon_chart.countall;
  ind_list_t*  elmin;
  fitnesst     minfit;
  int          regicount,vtlen,j,i=0;
  while(i < db) {
    vtlen=RaNdOm(ptovabb);
    for(iterator=taxon_chart.indvdls,j=0;j<vtlen;iterator=iterator->next,++j);
    minfit=(elmin=iterator)->e.fit();
    while(iterator != NULL) {
      regicount=iterator->spec->count;
      do {
        if(iterator->e.fit()<minfit) {
          elmin=iterator;
          minfit=elmin->e.fit(); };
        for(j=0;j<ptovabb && iterator!=NULL;iterator=iterator->next,++j);
      }
      while(iterator!=NULL && iterator->spec->count==regicount);
      if(iterator!=NULL) {
        minfit-=Gfitlowb;
        minfit=(minfit*iterator->spec->count)/regicount;
        minfit+=Gfitlowb; };
    };
    while(elmin!=NULL && !kill(elmin)) elmin=elmin->next;
    ++i;
  };
// may be that we haven't killed db individual
};

// -----------------------------------------------------------------------

void population::update_(taxon_chart_t* f) {
  f->indvdls->e.init(*task_);
  f->maxe=f->indvdls;
  f->fitsum=f->fitlowb=f->maxe->e.fit();
  iterstart(f);iterator=iterator->next;
  for(;!iterend();iterator=iterator->next) {
    iterator->e.init(*task_);
    if(iterator->e.fit() > f->maxe->e.fit()) f->maxe=iterator;
    else {
      if(iterator->e.fit() < f->fitlowb) f->fitlowb=iterator->e.fit();
      if(iterator->e.fit() < Gfitlowb) Gfitlowb=iterator->e.fit(); };
    f->fitsum+=iterator->e.fit(); };
  for(f=f->firstson;f != NULL;f=f->brother) update_(f);
};

// -----------------------------------------------------------------------

void population::update() {
  // updates fields fitsum,maxe,fitlowb,Gfitlowb and the indvdls
  taxon_chart.indvdls->e.init(*task_);
  Gfitlowb=taxon_chart.indvdls->e.fit();
  update_(&taxon_chart);
};

// -----------------------------------------------------------------------

void population::maximum(taxon_chart_t* f,signed char a) {
// if a=0 : maximum, if a=1 : maximum_all. default: a=0
  if(f->maxe->e.fit() > max_ind.fit()) max_ind=f->maxe->e;
  if(a==1 && f->firstson!=NULL)
    for(f=f->firstson;f!=NULL;f=f->brother) maximum(f,1);
};

// -----------------------------------------------------------------------

void population::sel_relax(unsigned int fok,taxon_chart_t* f) {
  // decrements fitlowb , and goes on to subspecies
  if(f==NULL) f=&taxon_chart;
  f->fitlowb-=(f->maxe->e.fit()-Gfitlowb)*fok; // this is a heuristic
  for(f=f->firstson;f != NULL;f=f->brother) sel_relax(fok,f);
};

// -----------------------------------------------------------------------

void population::dying_relax(unsigned int fok) {
  // finds out global lower bound
  Gfitlowb=taxon_chart.indvdls->e.fit();
  iterator=taxon_chart.indvdls;
  for(;iterator!=NULL;iterator=iterator->next)
    if(iterator->e.fit() < Gfitlowb) Gfitlowb=iterator->e.fit();
  // decrements fitlowb
  max_ind=taxon_chart.maxe->e;
  maximum(&taxon_chart,1);
  Gfitlowb-=(max_ind.fit()-Gfitlowb)*fok;
};

// -----------------------------------------------------------------------
// object qu for fusion

struct qu_elem {
  struct qu_elem* next;
  struct qu_elem* eloz;
  taxon_chart_t*  tc; };

class qu {
  struct qu_elem Q,*utolso,*seged;
public:
  qu() { utolso=&Q;Q.next=NULL;Q.eloz=NULL; };
  ~qu();
  signed char empty() { return utolso==&Q; };
  int  to_qu(taxon_chart_t* f) {
         seged=Q.next;
         if((Q.next=new struct qu_elem)==NULL) fewmem();
         Q.next->next=seged;
         Q.next->eloz=&Q;
         if(seged==NULL) utolso=Q.next; else Q.next->next->eloz=Q.next;
         Q.next->tc=f;return 0; };
  taxon_chart_t* from_qu() {
         Q.tc=utolso->tc;utolso=utolso->eloz;
         delete utolso->next;utolso->next=NULL;return Q.tc; };
};

qu::~qu() {
  for(seged=Q.next;seged!=NULL;seged=utolso) {
    utolso=seged->next;delete seged;
  };
};

// -----------------------------------------------------------------------

void population::unite(taxon_chart_t* mit,taxon_chart_t* hova) {
  ind_list_t    *seged,*eloz;
  taxon_chart_t *f,*pre;

  // copying ind.s of mit to hova

  if(hova->lastind->next!=mit->indvdls) {
    f=&taxon_chart;
    while(f != mit) {
      if(f->count != -1) pre=f;
      if(f->firstson != NULL) f=f->firstson;
      else
	if(f->brother != NULL) f=f->brother;
	else {
	  for(f=f->dad; f->brother==NULL; f=f->dad);
	  f=f->brother;
	};
    };
    eloz=pre->lastind;
    // eloz->next == mit->indvdls
    seged=hova->lastind->next;
    hova->lastind->next=mit->indvdls;
    eloz->next=mit->lastind->next;
    mit->lastind->next=seged; };

  // updating fields

  hova->rad_index=  // minimum of two rad_indexes
    (hova->rad_index<mit->rad_index ? hova->rad_index : mit->rad_index);
  hova->lastind=mit->lastind;
  seged=mit->indvdls;
  seged->spec=hova;
  while(seged!=mit->lastind) { seged=seged->next;seged->spec=hova; };
  hova->fitsum+=mit->fitsum;
  if(hova->maxe->e.fit() < mit->maxe->e.fit()) hova->maxe=mit->maxe;
  if(hova->fitlowb > mit->fitlowb) hova->fitlowb=mit->fitlowb;
  hova->count+=mit->count;
  for(f=hova;f!=NULL;f=f->dad) f->countall+=mit->count;
  for(f=mit;f!=NULL;f=f->dad) f->countall-=mit->count;
  mit->count=-1; // signes that the species is empty for later deletion
  --species_number;
};

// -----------------------------------------------------------------------

signed char population::del_holes(taxon_chart_t* f) {
  int            i;
  signed char    toroltunk=0;
  taxon_chart_t  *seged,*seged2,*ff;
  while(f->firstson != NULL && del_holes(f->firstson));
  if(f->firstson != NULL) {
    f=f->firstson;
    while(f->brother !=NULL) {
      while(del_holes(f->brother) && f->brother != NULL);
      if(f->brother != NULL) f=f->brother;
    };
    f=f->dad;
  };
  if(f->count==-1) {   // must be deleted
    toroltunk=1;
    ff=f;
    if(f->firstson == NULL)
      if(f == f->dad->firstson) f->dad->firstson=f->brother;
      else {
        for(seged=f->dad->firstson; seged->brother!=f; seged=seged->brother);
        seged->brother=f->brother;
      }
    else {
      if(f == f->dad->firstson) f->dad->firstson=f->firstson;
      else {
        for(seged=f->dad->firstson; seged->brother!=f; seged=seged->brother);
        seged->brother=f->firstson;
      };
      f->firstson->dad=f->dad;
      seged2=f->brother;
      do {
        seged=f->firstson->brother;i=0;
        for(; seged!=NULL; seged=seged->brother)
          { seged->dad=f->firstson;i+=seged->countall; };
        seged=f->firstson->brother;
        f->firstson->brother=seged2;
        f->firstson->countall+=i;
        seged2=seged; }
      while((f=f->firstson)->firstson != NULL);
      f->firstson=seged2;
    };
    delete ff;
  };
  return toroltunk;
};

// -----------------------------------------------------------------------

void population::unite_if_needed(taxon_chart_t* mit,double dopt,double d) {
  if(mit->count==-1) return;
  signed char    tovabb=1;
  qu             S;
  taxon_chart_t* f;
  S.to_qu(&taxon_chart);
  while(!S.empty()) {
    f=S.from_qu();
    do {
      if(f->count==-1) continue;
      if(f==mit) return;
      if((mit->maxe->e.fen()-f->maxe->e.fen() <= dopt) ||
                                          ( d >  0.5   &&
                                   mit->count <  4)) { // must be united
        if(mit->rad_index <= f->rad_index) unite(f,mit);
        else unite(mit,f);
        return;
      };
      if(f->firstson==mit) tovabb=0;
      if(tovabb && f->firstson!=NULL) S.to_qu(f->firstson);
    }
    while((f=f->brother)!=NULL); };
  return;
};

// -----------------------------------------------------------------------

void population::fusion(double dopt,double d,taxon_chart_t* f) {
  if(f->firstson==NULL) return;
  unite_if_needed(f->firstson,dopt,d);
  f=f->firstson;
  while(f->brother != NULL) {
    unite_if_needed(f->brother,dopt,d);
    f=f->brother; };
  for(f=f->dad->firstson; f != NULL; f=f->brother)
    if(f->firstson != NULL) fusion(dopt,d, f);
};
